/*
*********************************************
*											*
*		Astro TIME functions				*
*											*
*********************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "atime.h"
#include "dt_tbl.h"

/*
---------------------------------------------
		leap second table adr.
---------------------------------------------
*/
REAPSEC *pgdAT = NULL;

/*
#############################################
#											#
#			Julian day functions			#
#											#
#############################################
*/
/*
=============================================
=			date to julian					=
=============================================
*/
double julian(int nYear, int nMonth, double dDay)
{
	double da, db;
	double dJD;

	if (nMonth < 3){
		nMonth += 12;
		nYear--;
	}

	// Calc. body
	da = floor(nYear/100.0);
	db = 2.0 - da + floor(da / 4.0);
	dJD = floor(365.25 * nYear) + floor(30.6001 * (nMonth + 1)) + dDay + db + 1720994.5;

	if (dJD < 2299160.5){
		dJD -= db;
	}

	return dJD;
}

/*
=============================================
=			julian to date					=
=============================================
*/
void juian_date(double dJD, int *nYear, int *nMonth, double *dDay)
{
	int nY, nM;
	double da, db, dd;
	double dJDs = dJD + 0.5;
	double dJDi = floor(dJDs);
	double dJDf = dJDs - dJDi;

	if (dJDi > 2299160.0){
		da = floor((dJDi - 1867216.25) / 36524.25);
		dJDi += (1.0 + da - floor(da / 4.0));
	}

	db = dJDi + 1524.0;
	nY = (int)(floor((db - 122.1) / 365.25));
	dd = db - floor(nY * 365.25);
	nM = (int)(floor(dd / 30.6001));

	// Write Day
	*dDay = dd - floor(nM * 30.6001) + dJDf;

	// Write Month
	if (nM < 14){
		*nMonth = nM - 1;
	}else{
		*nMonth = nM - 13;
	}

	// Write Year
	if (*nMonth > 2){
		*nYear = nY - 4716;
	}else{
		*nYear = nY - 4715;
	}
}

/*
=============================================
=			julian to period day			=
=============================================
*/
int jd_period(double dJD, int nPerioDay)
{
	int nJD = (int)floor(dJD + 0.5);
	int nP = nJD % nPerioDay;
	if (nP < 0){
		nP += nPerioDay;
	}

	return nP;
}



/*
#############################################
#											#
#			UTC <-> TAI or TT functions		#
#											#
#############################################
*/
/*
=============================================
=		TT(TT-UTC) function (1972`)		=
=============================================
*/
/*
---------------------------------------------
-		set TT table adr.					-
---------------------------------------------
*/
void set_dtt_table(REAPSEC *pdTbl)
{
	pgdAT = pdTbl;
}

/*
---------------------------------------------
-		get TT from UTC(Date only)			-
-											-
-entry: double dJD = JD_UTC(date only)		-
-exit:	return double T(dayunit)			-
---------------------------------------------
*/
double get_dtt_1972(double dJD_utc)
{
	REAPSEC *pRTbl;

	pRTbl = pgdAT;
	while( dJD_utc < pRTbl->dJD){
		pRTbl++;
	}
	return (pRTbl->dTime);
}

/*
---------------------------------------------
-		get TT from TT						-
-											-
-entry: double dJD_tt = JD_TT				-
-		int *pn60s = write 60s flag			-
-											-
-exit:	return double T(dayunit)			-
-		*pn60s = 0:No 60s 1:60s				-
---------------------------------------------
*/
double get_dtt_tt_1972(double dJD_tt, int *pn60s)
{
	REAPSEC *ptbl;
	double dDTT; 

	/* search table */
	ptbl = pgdAT;
	while( dJD_tt < (ptbl->dJD + ptbl->dTime)){
		ptbl++;
	}
	dDTT = ptbl->dTime;

	/* check 60s data */
	*pn60s = 0;
	if (ptbl != pgdAT){
		ptbl--;
		if ((dJD_tt - dDTT) >= ptbl->dJD){
			/* 60s unit */
			*pn60s = 1;
		}
	}
	
	return dDTT;
}

/*
=============================================
=		get T table data(1860`1960)		=
=											=
=entry: double dJD = JD_UTC					=
=exit:	return double T(dayunit)			=
=============================================
*/
double get_dt_table(double dJD)
{
	REAPSEC *ptbl;
	double dTemp1, dTemp2;
	int i;

	/* about jd to year */
	i = (int)((dJD - 2400410.5) / GREGOR_YEARDAY) - 1;
	if (i < 0){
		i = 0;
	}

	/* scan table */
	ptbl = &tTimesDT[i];
	while (dJD >= ptbl->dJD){
		ptbl++;
	}
	ptbl--;

	/* calc. interpolation */
	dTemp1 = (ptbl + 1)->dJD - ptbl->dJD;
	dTemp2 = (ptbl + 1)->dTime - ptbl->dTime;
	dTemp2 /= dTemp1;

	dTemp1 = ptbl->dTime + (dJD - ptbl->dJD) * dTemp2;

	return (dTemp1 / DAYSEC);
}


/*
=============================================
=			get T data	 approxmate			=
=											=
=entry: double dJD = JD_UTC					=
=exit:	return double T(dayunit)			=
=============================================
*/
double get_dt_approx(double dJD)
{
	double **pDATbl;
	double *pDTbl;
	double dT, dY;
	double dDT;

	/* scan table */
	pDATbl = pDT;
	while(*pDATbl != NULL){
		if(dJD >= **pDATbl){
			break;
		}
		pDATbl++;
	}
	if (*pDATbl == NULL){
		/* pre -500 */
		dY = (dJD - 2385800.5) / (JULIAN_YEARDAY*100);
		dDT = -20.0 + 32.0 * dY * dY;
		return (dDT / DAYSEC);
	}
	
	/* calc Y */
	pDTbl = *pDATbl;
	pDTbl++;
	dY = dJD - *pDTbl++;
	if (*pDTbl != 0.0){
		dY *= *pDTbl;
	}
	pDTbl++;

	/* calc Cn */
	dDT = *pDTbl++;
	dT = dY;
	while(*pDTbl != 0.0){
		dDT += (*pDTbl++ * dT);
		dT *= dY;
	}

	return (dDT / DAYSEC);
}

/*
=============================================
=		get T 	(all year)					=
=											=
=entry: double dJD = JD_UTC					=
=exit:	return double T(dayunit)			=
=============================================
*/
double get_dt(double dJD)
{
	/* cal. parameter table */
	if (dJD >= 2441317.5){
		/* (1971`) */
		return get_dtt_1972(dJD);
	}
	if (dJD >= 2400410.5){
		/* 1860`1971 */
		return get_dt_table(dJD);
	}

	/* other NASA */
	return get_dt_approx(dJD);
}

/*
=============================================
=			UTC to TAI or TT				=
=											=
=entry: DATETIME *pUTC = UTC date & time	=
=		double *pdJD_tt = TAI store adr.	=
=		REAPSEC pdAT = dAT Table adr.		=
=											=
=exit:	return double TT-UTC sec. unit		=
=		pdJD_tai in TAI						=
=============================================
*/
double utc_tt(DATETIME *pUTC, double *pdJD_tt)
{
	double dJD;
	double ddAT;
	double dTimeDay;

	/*	make date to julian	& time_day */
	dJD = julian(pUTC->nYear, pUTC->nMonth, (double)pUTC->nDay);
	dTimeDay = pUTC->nHour / 24.0 + pUTC->nMin / 1440.0 + pUTC->dSec / DAYSEC;

	/* get delta AT	*/
	if(dJD >= 2441317.5){
		ddAT = get_dtt_1972(dJD);
	}else{
		ddAT = get_dt(dJD + dTimeDay);
	}

	/* add deltaAT */
	*pdJD_tt = dJD + dTimeDay + ddAT;
	return (ddAT * DAYSEC);
}

/*
=============================================
=			TT to UTC						=
=											=
=entry: double *pdJD_tt = TT data			=
=		DATETIME *pUTC = UTC store adr		=
=		REAPSEC pdAT = dAT Table adr.		=
=											=
=exit:	return double TAI-UTC sec. unit		=
=		pUTC in UTC date & time				=
=============================================
*/
double tt_utc(double dJD_tt, DATETIME *pUTC)
{
	double dJD;
	double dDay;
	double ddAT;
	int n60sec = 0;

	if (dJD_tt >= (2441317.5 + (10 + DTAI_TT) / DAYSEC)){
		ddAT = get_dtt_tt_1972(dJD_tt, &n60sec);
		if (n60sec != 0){
			dJD_tt -= (1.0/DAYSEC);
		}
	}else{
		/* `1971 */
		ddAT = 0;
		do{
			dJD = ddAT;
			ddAT = get_dt(dJD_tt - dJD);
		}while(fabs(dJD - ddAT) > (0.1 / DAYSEC));
	}

	/*	julian to date */
	dJD = dJD_tt - ddAT;
	juian_date(dJD, &pUTC->nYear, &pUTC->nMonth, &dDay);
	pUTC->nDay = (int)floor(dDay);
	dayfrac_time(dDay - pUTC->nDay, &pUTC->nHour, &pUTC->nMin, &pUTC->dSec);

	/* make 60s */
	pUTC->dSec += n60sec;

	return (ddAT * DAYSEC);
}

/*
#############################################
#											#
#		other common functions				#
#											#
#############################################
*/
/*
=============================================
=		daytime to time structer			=
=============================================
*/
void dayfrac_time(double dDayFrac, int *pnHour, int *pnMin, double *pdSec)
{
	double dTemp;

	dTemp = dDayFrac * DAYHOUR;
	*pnHour = (int)floor(dTemp);
	dTemp -= *pnHour;

	dTemp *= 60.0;
	*pnMin = (int)floor(dTemp);
	dTemp -= *pnMin;

	*pdSec = dTemp * 60.0;
}

